/* 
 * /bin/su --
 * Heavily modified for use on HP-UX.
 * rthomas@pamd.cig.mot.com, Service Team PAMD, ITS, Motorola CIG
 * 04.13.95
 *
 * This bit o' code has been modified in the following ways:
 * 1. Acts in accordance with the 'su' man page for HP-UX 9.xx
 * 2. Allows for additional logging of su attempts via syslog
 * 3. Cleaned up a bit... but I left the heinous "goto" 8-)
 * 4. To su to root, any user, *INCLUDING* root, must properly give the
 *    root password.  This is necessary because of a hole in HP-UX boot-up
 * 5. SULOG has a mode of 600.  Safe from prying non-root eyes.
 *
 * This flavour of su should be more secure, as it allows for the additional
 * logging and prevention of single-user su without giving the root password.
 *
 * If SULOG is defined, all attempts to su to uid 0 are logged in that spot.
 * If SYSLOG is defined, all attempts to su to uid 0 are logged in SYSLOG,
 * logging level XXXXX.
 *
 * Please setup all definitions in the Makefile.  That is what it's there
 * for...
 *
 * usage: su [-] [name] [-c command] [-r]
 * -  = changes environment
 * -c = execs following command, exits
 * -r = launches a restricted shell
 */

#ifndef lint
static char sccsid[] = "@(#)su.c v1.0 13 APR 1995 Rob Thomas robt@cymru.com";
#endif

#include <stdio.h>
#include <pwd.h>
#include <time.h>
#include <signal.h>
#include <sys/stat.h> 			/* added for chmod change */
#include <syslog.h>

#define PATH	"PATH=:/bin:/usr/bin"
#define SUPATH	"PATH=/bin:/etc:/usr/bin"
#define ELIM 128

long time();
void pause();
struct	passwd *pwd, *getpwnam();
struct	passwd *opwd, *getpwuid();
struct	tm *localtime();
char	*malloc(), *strcpy();
char	*getpass(), *ttyname(), *strrchr();
char	*shell = "/bin/sh";
char	su[16] = "su";
char	homedir[64] = "HOME=";
char	logname[20] = "LOGNAME=";
char	*path = PATH;
char	*supath = SUPATH;
char	*envinit[ELIM];
extern	char **environ;
char	*ttyn;

int
main(argc, argv)
	int	argc;
	char	**argv;
{
	char	*nptr, *password;
	char	*pshell = shell;
	char	*runcmd;
	int	badsw = 0;
	int	eflag = 0;
	int	uid, gid, wutmyuid;
	char	*dir, *shprog, *name, *wutmynam;
	char	option;

	wutmyuid = getuid();
	if((opwd = getpwuid(wutmyuid)) == NULL) {
		fprintf(stderr, "su: Unknown id: %s\n", wutmyuid);
		exit(1);
	}

	wutmynam = strcpy(malloc(strlen(opwd->pw_name)+1), opwd->pw_name);

	if (argc > 1 && *argv[1] == '-') {
		eflag++;
		argv++;
		argc--;
	}
	nptr = (argc > 1)? argv[1]: "root";
	if((pwd = getpwnam(nptr)) == NULL) {
		fprintf(stderr,"su: Unknown id: %s\n",nptr);
		exit(1);
	}
	uid = pwd->pw_uid;
	gid = pwd->pw_gid;
	dir = strcpy(malloc(strlen(pwd->pw_dir)+1),pwd->pw_dir);
	shprog = strcpy(malloc(strlen(pwd->pw_shell)+1),pwd->pw_shell);
	name = strcpy(malloc(strlen(pwd->pw_name)+1),pwd->pw_name);
	if((ttyn=ttyname(0))==NULL)
		if((ttyn=ttyname(1))==NULL)
			if((ttyn=ttyname(2))==NULL)
				ttyn="/dev/tty??";
/*
 * Added check to make sure no one can just "su" to root without a
 * password -- NOT EVEN ROOT!
 */
	if((pwd->pw_passwd[0] == '\0' || (uid != 0 && getuid() == 0)))
		goto ok;
	password = getpass("Password:");
	if(badsw || (strcmp(pwd->pw_passwd, crypt(password, pwd->pw_passwd)) != 0))	{
#ifdef SULOG
		sulog(SULOG, nptr, 0, uid, wutmynam);
#endif
		fprintf(stderr,"su: Sorry\n");
		exit(2);
	}
ok:
	endpwent();
#ifdef SULOG
		sulog(SULOG, nptr, 1, uid, wutmynam);
#endif
	initgroups(name, gid);	/* get group list before we are user */
	if((setgid(gid) != 0) || (setuid(uid) != 0)) {
		printf("Invalid ID\n");
		exit(2);
	}
	argv++;
	argc--;
	if (eflag) {
		strcat(homedir, dir);
		strcat(logname, name);
		envinit[2] = logname;
		chdir(dir);
		envinit[0] = homedir;
		if (uid == 0)
			envinit[1] = supath;
		else	envinit[1] = path;
		envinit[3] = NULL;
		environ = envinit;
		strcpy(su, "-su");
	}

	while ((option = getopt(argc, argv, "c:r")) != -1)
		switch (option) {
			case 'c':
				sprintf(su, "-c");
				runcmd = optarg;
				pshell = shprog;
				exit(execlp(pshell, pshell, su, runcmd, 0));
				break;
			case 'r':
				strcpy(su, eflag ? "-" : "");
				strcat(su, "rsh");
				exit(execlp(RSH, su, 0));
				break;
			case '?':
				fprintf(stderr, "su: What is %c?\n",
					optopt);
				exit(3);
				break;
			}	

	if(shprog[0] != '\0') {
		pshell = shprog;
		strcpy(su, eflag ? "-" : "");
		strcat(su, strrchr(pshell,'/') + 1);
	}
	execl(pshell, su, 0);

	fprintf(stderr,"su: No shell\n");
	exit(4);
}

int
envalt()
{
	int	i, pset=0, j=0;
	char	**eptr=environ;

	for(i=0;i<130;i++)
	{
		if(*eptr == (char *)0) break;
		if(strncmp(*eptr,"PATH=",5)==0)
		{
			envinit[i-j]=supath;
			pset++;
		}
		else if(strncmp(*eptr,"PS1=",4)==0)
			j++;
		else
  			envinit[i-j] = *eptr;

		eptr++;
	}

	if(!pset)
	{
		i = (i < ELIM-2) ? i : ELIM-2;
		envinit[i++]=supath;
	}
	envinit[((i < ELIM-1) ? i : ELIM-1)]=(char *)0;

	environ = envinit;
}

int
sulog(where, towho, how, uid, wutmynam)
	char	*where, *towho, *wutmynam;
	int	how;
	int	uid;
{
/*
 * Added chmod change so that logfile is mode 600...
 * Added syslog functionality.  Good info...
 */
	FILE	*logf;
	long	now;
	struct	tm *tmp;

	now = time(0);
	tmp = localtime(&now);
	if((logf=fopen(where,"a")) != NULL) {
		fprintf(logf,"SU %.2d/%.2d %.2d:%.2d %c %s %s-%s\n",
		  tmp->tm_mon+1,tmp->tm_mday,tmp->tm_hour,tmp->tm_min,
		  how?'+':'-',(strrchr(ttyn,'/')+1),wutmynam,towho);
		fclose(logf);
		chmod(logf, S_IRUSR|S_IWUSR);
	}

#ifdef SYSLOG
	if (0 == uid) {
		openlog("su", LOG_PID, FACCODE);
		syslog(PRIORITY, "SU %.2d/%.2d %.2d:%.2d %c %s %s-%s\n",
		  tmp->tm_mon+1,tmp->tm_mday,tmp->tm_hour,tmp->tm_min,
		  how?'+':'-',(strrchr(ttyn,'/')+1),wutmynam,towho);
		closelog();
	}
#endif
}
